﻿Imports System.IO
Imports System.Drawing

Public Class Form1
    Private Sub ButtonLoad_Click(sender As Object, e As EventArgs) Handles ButtonLoad.Click
        Using ofd As New OpenFileDialog
            ofd.Filter = "ILBM Files (*.iff;*.lbm)|*.iff;*.lbm"
            If ofd.ShowDialog() = DialogResult.OK Then
                Dim bmp As Bitmap = LoadILBM(ofd.FileName)
                If bmp IsNot Nothing Then
                    PictureBox1.Image = bmp
                Else
                    MessageBox.Show("Failed to load ILBM file.")
                End If
            End If
        End Using
    End Sub

    Private Function LoadILBM(filePath As String) As Bitmap
        Using fs As New FileStream(filePath, FileMode.Open, FileAccess.Read)
            Using br As New BinaryReader(fs)
                ' Read FORM header
                Dim formID = New String(br.ReadChars(4))
                Dim formSize = ReadUInt32BE(br)
                Dim ilbmID = New String(br.ReadChars(4))

                If formID <> "FORM" OrElse ilbmID <> "ILBM" Then Return Nothing

                Dim width As Integer = 0
                Dim height As Integer = 0
                Dim nPlanes As Integer = 0
                Dim palette(255) As Color
                Dim bodyData As Byte() = Nothing
                Dim compression As Integer = 0 ' default to uncompressed

                ' Parse chunks
                While br.BaseStream.Position < br.BaseStream.Length
                    Dim chunkStart = br.BaseStream.Position
                    Dim chunkID = System.Text.Encoding.ASCII.GetString(br.ReadBytes(4))
                    Dim chunkSize = ReadUInt32BE(br)
                    Dim chunkDataStart = br.BaseStream.Position

                    Select Case chunkID
                        Case "BMHD"
                            width = ReadInt16BE(br)
                            height = ReadInt16BE(br)
                            br.BaseStream.Seek(4, SeekOrigin.Current) ' x, y
                            nPlanes = br.ReadByte()
                            br.ReadByte()
                            compression = br.ReadByte()
                            br.BaseStream.Seek(7, SeekOrigin.Current) ' masking, compression, etc.

                        Case "CMAP"
                            For i = 0 To (chunkSize \ 3) - 1
                                Dim r = br.ReadByte()
                                Dim g = br.ReadByte()
                                Dim b = br.ReadByte()
                                palette(i) = Color.FromArgb(r, g, b)
                            Next

                        Case "BODY"
                            Dim rawData = br.ReadBytes(CInt(chunkSize))
                            If compression = 1 Then
                                bodyData = DecompressByteRun1(rawData)
                            Else
                                bodyData = rawData
                            End If
                        Case Else
                            ' Unknown chunk, skip it
                            br.BaseStream.Seek(chunkSize, SeekOrigin.Current)
                    End Select

                    ' Always move to next chunk (even-aligned)
                    Dim nextChunkPos = chunkDataStart + chunkSize
                    If chunkSize Mod 2 = 1 Then nextChunkPos += 1
                    br.BaseStream.Position = nextChunkPos
                End While

                If bodyData Is Nothing OrElse width = 0 OrElse height = 0 Then Return Nothing

                Return ConvertToBitmap(width, height, nPlanes, palette, bodyData)
            End Using
        End Using
    End Function

    Private Function ConvertToBitmap(width As Integer, height As Integer, planes As Integer, palette() As Color, body() As Byte) As Bitmap
        Dim bmp As New Bitmap(width, height)
        Dim rowSize = ((width + 15) \ 16) * 2 ' padded row size in bytes
        Dim stride = rowSize * planes
        Dim offset = 0

        For y = 0 To height - 1
            Dim bitplanes(planes - 1)() As Byte
            For p = 0 To planes - 1
                bitplanes(p) = body.Skip(offset + p * rowSize).Take(rowSize).ToArray()
            Next
            offset += stride

            For x = 0 To width - 1
                Dim colorIndex As Integer = 0
                For p = 0 To planes - 1
                    Dim byteIndex = x \ 8
                    Dim bit = 7 - (x Mod 8)
                    If (bitplanes(p)(byteIndex) And (1 << bit)) <> 0 Then
                        colorIndex = colorIndex Or (1 << p)
                    End If
                Next
                bmp.SetPixel(x, y, palette(colorIndex Mod palette.Length))
            Next
        Next

        Return bmp
    End Function

    ' ByteRun1 decompression
    Private Function DecompressByteRun1(data As Byte()) As Byte()
        Dim output As New List(Of Byte)
        Dim i As Integer = 0

        While i < data.Length
            Dim n As Integer = data(i)
            If n > 127 Then n -= 256 ' Convert to SByte range
            i += 1

            If n >= 0 AndAlso n <= 127 Then
                ' Literal run
                If i + n + 1 <= data.Length Then
                    output.AddRange(data.Skip(i).Take(n + 1))
                End If
                i += n + 1
            ElseIf n >= -127 AndAlso n <= -1 Then
                ' Repeat next byte (-n + 1 times)
                If i < data.Length Then
                    Dim value As Byte = data(i)
                    For j = 0 To -n
                        output.Add(value)
                    Next
                    i += 1
                End If
            ElseIf n = -128 Then
                ' No-op
            End If
        End While

        Return output.ToArray()
    End Function

    ' Big-endian read helpers
    Private Function ReadUInt16BE(br As BinaryReader) As UShort
        Dim bytes = br.ReadBytes(2)
        Array.Reverse(bytes)
        Return BitConverter.ToUInt16(bytes, 0)
    End Function

    Private Function ReadInt16BE(br As BinaryReader) As Short
        Dim bytes = br.ReadBytes(2)
        Array.Reverse(bytes)
        Return BitConverter.ToInt16(bytes, 0)
    End Function

    Private Function ReadUInt32BE(br As BinaryReader) As UInteger
        Dim bytes = br.ReadBytes(4)
        Array.Reverse(bytes)
        Return BitConverter.ToUInt32(bytes, 0)
    End Function
End Class

